"""Utility functions for experiments."""

__author__ = ["MatthewMiddlehurst"]
__all__ = [
    "assign_gpu",
    "timing_benchmark",
    "estimator_attributes_to_file",
]

import os
import time
from collections.abc import Sequence

import gpustat
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils import check_random_state


def _results_present(path, estimator, dataset, resample_id=None, split="TEST"):
    """Check if results are present already."""
    resample_str = "Results" if resample_id is None else f"Resample{resample_id}"
    path = f"{path}/{estimator}/Predictions/{dataset}/"

    if split == "BOTH":
        full_path = f"{path}test{resample_str}.csv"
        full_path2 = f"{path}train{resample_str}.csv"

        if os.path.exists(full_path) and os.path.exists(full_path2):
            return True
    else:
        if split is None or split == "" or split == "NONE":
            full_path = f"{path}{resample_str.lower()}.csv"
        elif split == "TEST":
            full_path = f"{path}test{resample_str}.csv"
        elif split == "TRAIN":
            full_path = f"{path}train{resample_str}.csv"
        else:
            raise ValueError(f"Unknown split value: {split}")

        if os.path.exists(full_path):
            return True

    return False


def _check_existing_results(
    results_path,
    estimator_name,
    dataset,
    resample_id,
    overwrite,
    build_test_file,
    build_train_file,
):
    """Check if results are present already and if they should be overwritten."""
    if not overwrite:
        resample_str = "Result" if resample_id is None else f"Resample{resample_id}"

        if build_test_file:
            full_path = (
                f"{results_path}/{estimator_name}/Predictions/{dataset}/"
                f"/test{resample_str}.csv"
            )

            if os.path.exists(full_path):
                build_test_file = False

        if build_train_file:
            full_path = (
                f"{results_path}/{estimator_name}/Predictions/{dataset}/"
                f"/train{resample_str}.csv"
            )

            if os.path.exists(full_path):
                build_train_file = False

    return build_test_file, build_train_file


def assign_gpu(set_environ=False):  # pragma: no cover
    """Assign a GPU to the current process.

    Looks at the available Nvidia GPUs and assigns the GPU with the lowest used memory.

    Parameters
    ----------
    set_environ : bool
        Set the CUDA_DEVICE_ORDER environment variable to "PCI_BUS_ID" anf the
        CUDA_VISIBLE_DEVICES environment variable to the assigned GPU.

    Returns
    -------
    gpu : int
        The GPU assigned to the current process.
    """
    stats = gpustat.GPUStatCollection.new_query()
    pairs = [
        [
            gpu.entry["index"],
            float(gpu.entry["memory.used"]) / float(gpu.entry["memory.total"]),
        ]
        for gpu in stats
    ]

    gpu = min(pairs, key=lambda x: x[1])[0]

    if set_environ:
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu)

    return gpu


def timing_benchmark(num_arrays=1000, array_size=20000, random_state=None):
    """
    Measures the time taken to sort a given number of numpy arrays of a specified size.

    Returns the time taken in milliseconds.

    Parameters
    ----------
    num_arrays: int, default=1000
        Number of arrays to generate and sort.
    array_size: int, default=20000
        Size of each numpy array to be sorted.
    random_state: int, RandomState instance or None, default=None
        If `int`, random_state is the seed used by the random number generator;
        If `RandomState` instance, random_state is the random number generator;
        If `None`, the random number generator is the `RandomState` instance used
        by `np.random`.

    Returns
    -------
    time_taken: int
        Time taken to sort the arrays in milliseconds.
    """
    if random_state is None:
        random_state = check_random_state(0)
    elif isinstance(random_state, (int, np.random.RandomState)):
        random_state = check_random_state(random_state)
    else:
        raise ValueError("random_state must be an int, RandomState instance or None")

    total_time = 0
    for _ in range(num_arrays):
        array = random_state.rand(array_size)
        start_time = time.time()
        np.sort(array)
        end_time = time.time()
        total_time += end_time - start_time

    return int(round(total_time * 1000))


def estimator_attributes_to_file(
    estimator, dir_path, estimator_name=None, max_depth=np.inf, max_list_shape=np.inf
):
    """Write the attributes of an estimator to file(s).

    Write the attributes of an estimator to file at a given directory. The function
    will recursively write the attributes of any estimators or non-string sequences
    containing estimators found in the attributes of the input estimator to spearate
    files.

    Parameters
    ----------
    estimator : estimator instance
        The estimator to write the attributes of.
    dir_path : str
        The directory to write the attribute files to.
    estimator_name : str or None, default=None
        The name of the estimator. If None, the name of the estimator class will be
        used.
    max_depth : int, default=np.inf
        The maximum depth to go when recursively writing attributes of estimators.
    max_list_shape : int, default=np.inf
        The maximum shape of a list to write when recursively writing attributes of
        contained estimators. i.e. for 0, no estimators contained in lists will be
        written, for 1, only estimators contained in 1-dimensional lists or the top
        level of a list will be written.
    """
    estimator_name = (
        estimator.__class__.__name__ if estimator_name is None else estimator_name
    )
    _write_estimator_attributes_recursive(
        estimator, dir_path, estimator_name, 0, max_depth, max_list_shape
    )


def _write_estimator_attributes_recursive(
    estimator, dir_path, file_name, depth, max_depth, max_list_shape
):
    if depth > max_depth:
        return

    path = f"{dir_path}/{file_name}.txt"
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w") as file:
        for attr in estimator.__dict__:
            value = getattr(estimator, attr)
            file.write(f"{attr}: {value}\n")

            if depth + 1 <= max_depth:
                if isinstance(value, BaseEstimator):
                    new_dir_path = f"{dir_path}/{attr}/"
                    file.write(f"    See {new_dir_path}{attr}.txt for more details\n")
                    _write_estimator_attributes_recursive(
                        value, new_dir_path, attr, depth + 1, max_depth, max_list_shape
                    )
                elif _is_non_string_sequence(value):
                    _write_list_attributes_recursive(
                        value,
                        file,
                        dir_path,
                        attr,
                        depth + 1,
                        max_depth,
                        1,
                        max_list_shape,
                    )


def _write_list_attributes_recursive(
    it, file, dir_path, file_name, depth, max_depth, shape, max_list_shape
):
    if shape > max_list_shape:
        return

    for idx, item in enumerate(it):
        if isinstance(item, BaseEstimator):
            new_dir_path = f"{dir_path}/{file_name}_{idx}/"
            file.write(
                f"    See {new_dir_path}{file_name}_{idx}.txt for more details\n"
            )
            _write_estimator_attributes_recursive(
                item,
                new_dir_path,
                f"{file_name}_{idx}",
                depth,
                max_depth,
                max_list_shape,
            )
        elif _is_non_string_sequence(item):
            _write_list_attributes_recursive(
                item,
                file,
                dir_path,
                f"{file_name}_{idx}",
                depth,
                max_depth,
                shape + 1,
                max_list_shape,
            )


def _is_non_string_sequence(obj):
    return isinstance(obj, Sequence) and not isinstance(obj, (str, bytes, bytearray))



"""Functions for running experiments using cross-validation."""

__author__ = ["MatthewMiddlehurst"]
__all__ = [
    "classification_cross_validation",
    "classification_cross_validation_folds",
    "regression_cross_validation",
    "regression_cross_validation_folds",
]

import numpy as np
from sklearn.model_selection import KFold, StratifiedKFold

from tsml_eval.experiments.experiments import (
    run_classification_experiment,
    run_regression_experiment,
)


def classification_cross_validation(
    X,
    y,
    estimator,
    results_path,
    cv=None,
    fold_ids=None,
    row_normalise=False,
    classifier_name=None,
    dataset_name="N/A",
    build_test_file=True,
    build_train_file=False,
    ignore_custom_train_estimate=False,
    attribute_file_path=None,
    att_max_shape=0,
    benchmark_time=True,
):
    """Run a classification experiment using cross-validation.

    Parameters
    ----------
    X : array-like
        Feature data.
    y : array-like
        Target labels.
    estimator : object
        The classifier to be evaluated.
    results_path : str
        Path to save results.
    cv : object, optional
        Cross-validation strategy. If None, 10-fold cross-validation will be used.
    fold_ids : list, optional
        List of fold ids to run. If None, all folds will be run.
        row_normalise : bool, default=False
        Whether to normalise the data rows (time series) prior to fitting and
        predicting.
    classifier_name : str or None, default=None
        Name of classifier used in writing results. If None, the name is taken from
        the classifier.
    dataset_name : str, default="N/A"
        Name of dataset.
    build_test_file : bool, default=True:
        Whether to generate test files or not. If the classifier can generate its own
        train probabilities, the classifier will be built but no file will be output.
    build_train_file : bool, default=False
        Whether to generate train files or not. If true, it performs a 10-fold
        cross-validation on the train data and saves. If the classifier can produce its
        own estimates, those are used instead.
    ignore_custom_train_estimate : bool, default=False
        todo
    attribute_file_path : str or None, default=None
        todo (only test)
    att_max_shape : int, default=0
        todo
    benchmark_time : bool, default=True
        Whether to benchmark the hardware used with a simple function and write the
        results. This will typically take ~2 seconds, but is hardware dependent.
    """
    if cv is None:
        cv_size = 10
        _, counts = np.unique(y, return_counts=True)
        min_class = np.min(counts)
        if min_class < cv_size:
            cv_size = min_class
            if cv_size < 2:
                raise ValueError(
                    "All classes must have at least 2 values to run the "
                    "default cross-validation."
                )
        cv = StratifiedKFold(n_splits=cv_size, shuffle=True, random_state=0)
    folds = list(cv.split(X, y))

    if fold_ids is not None:
        if isinstance(fold_ids, int):
            fold_ids = [fold_ids]
        if len(fold_ids) > len(folds) or max(fold_ids) >= len(folds):
            raise ValueError("Fold ids are to large for the number of folds.")
    else:
        fold_ids = list(range(len(folds)))

    for fold, (train, test) in enumerate(folds):
        if fold not in fold_ids:
            continue

        run_classification_experiment(
            X[train],
            y[train],
            X[test],
            y[test],
            estimator,
            results_path,
            row_normalise=row_normalise,
            classifier_name=classifier_name,
            dataset_name=dataset_name,
            resample_id=fold,
            build_test_file=build_test_file,
            build_train_file=build_train_file,
            ignore_custom_train_estimate=ignore_custom_train_estimate,
            attribute_file_path=attribute_file_path,
            att_max_shape=att_max_shape,
            benchmark_time=benchmark_time,
        )


def classification_cross_validation_folds(X, y, cv=None):
    """Get the folds for a classification cross-validation experiment.

    Parameters
    ----------
    X : array-like
        Feature data.
    y : array-like
        Target labels.
    cv : object, optional
        Cross-validation strategy. If None, 10-fold cross-validation will be used.
    """
    if cv is None:
        cv_size = 10
        _, counts = np.unique(y, return_counts=True)
        min_class = np.min(counts)
        if min_class < cv_size:
            cv_size = min_class
            if cv_size < 2:
                raise ValueError(
                    "All classes must have at least 2 values to run the "
                    "default cross-validation."
                )
        cv = StratifiedKFold(n_splits=10, shuffle=True, random_state=0)
    return list(cv.split(X, y))


def regression_cross_validation(
    X,
    y,
    estimator,
    results_path,
    cv=None,
    fold_ids=None,
    row_normalise=False,
    regressor_name=None,
    dataset_name="",
    build_test_file=True,
    build_train_file=False,
    ignore_custom_train_estimate=False,
    attribute_file_path=None,
    att_max_shape=0,
    benchmark_time=True,
):
    """Run a regression experiment using cross-validation.

    Parameters
    ----------
    X : array-like
        Feature data.
    y : array-like
        Target labels.
    estimator : object
        The regressor to be evaluated.
    results_path : str
        Path to save results.
    cv : object, optional
        Cross-validation strategy. If None, 10-fold cross-validation will be used.
    fold_ids : list, optional
        List of fold ids to run. If None, all folds will be run.
        row_normalise : bool, default=False
        Whether to normalise the data rows (time series) prior to fitting and
        predicting.
    regressor_name : str or None, default=None
        Name of regressor used in writing results. If None, the name is taken from
        the regressor.
    dataset_name : str, default="N/A"
        Name of dataset.
    build_test_file : bool, default=True:
        Whether to generate test files or not. If the regressor can generate its own
        train predictions, the classifier will be built but no file will be output.
    build_train_file : bool, default=False
        Whether to generate train files or not. If true, it performs a 10-fold
        cross-validation on the train data and saves. If the regressor can produce its
        own estimates, those are used instead.
    ignore_custom_train_estimate : bool, default=False
        todo
    attribute_file_path : str or None, default=None
        todo (only test)
    att_max_shape : int, default=0
        todo
    benchmark_time : bool, default=True
        Whether to benchmark the hardware used with a simple function and write the
        results. This will typically take ~2 seconds, but is hardware dependent.
    """
    if cv is None:
        cv = KFold(n_splits=10, shuffle=True, random_state=0)
    folds = list(cv.split(X, y))

    if fold_ids is not None:
        if isinstance(fold_ids, int):
            fold_ids = [fold_ids]
        if len(fold_ids) > len(folds) or max(fold_ids) >= len(folds):
            raise ValueError("Fold ids are to large for the number of folds.")
    else:
        fold_ids = list(range(len(folds)))

    for fold, (train, test) in enumerate(folds):
        if fold not in fold_ids:
            continue

        run_regression_experiment(
            X[train],
            y[train],
            X[test],
            y[test],
            estimator,
            results_path,
            row_normalise=row_normalise,
            regressor_name=regressor_name,
            dataset_name=dataset_name,
            resample_id=fold,
            build_test_file=build_test_file,
            build_train_file=build_train_file,
            ignore_custom_train_estimate=ignore_custom_train_estimate,
            attribute_file_path=attribute_file_path,
            att_max_shape=att_max_shape,
            benchmark_time=benchmark_time,
        )


def regression_cross_validation_folds(X, y, cv=None):
    """Get the folds for a regression cross-validation experiment.

    Parameters
    ----------
    X : array-like
        Feature data.
    y : array-like
        Target labels.
    cv : object, optional
        Cross-validation strategy. If None, 10-fold cross-validation will be used.
    """
    if cv is None:
        cv = KFold(n_splits=10, shuffle=True, random_state=0)
    return list(cv.split(X, y))



"""Distance-based clustering tests."""

__all__ = ["_DISTANCE_TEST_RESULTS_PATH"]

import os
from pathlib import Path

_DISTANCE_TEST_RESULTS_PATH = (
    os.path.dirname(Path(__file__).parent.parent.parent.parent.parent)
    + "/test_output/distance_clustering/"
)



"""Set regressor function."""

__author__ = ["TonyBagnall", "MatthewMiddlehurst"]

import numpy as np

from tsml_eval.utils.functions import str_in_nested_list

convolution_based_regressors = [
    ["rocketregressor", "rocket"],
    ["minirocket", "minirocketregressor"],
    ["multirocket", "multirocketregressor"],
    ["hydraregressor", "hydra"],
    ["multirockethydraregressor", "multirockethydra", "multirocket-hydra"],
]
deep_learning_regressors = [
    ["cnnregressor", "cnn"],
    ["fcnregressor", "fcnn", "fcn"],
    ["mlpregressor", "mlp"],
    ["encoderregressor", "encoder"],
    ["tapnetregressor", "tapnet"],
    ["resnetregressor", "resnet"],
    ["individualinceptionregressor", "singleinception", "individualinception"],
    ["inceptiontimeregressor", "inception", "inceptiontime"],
    ["h-inceptiontimeregressor", "h-inceptiontime"],
    ["litetimeregressor", "litetime"],
]
distance_based_regressors = [
    "1nn-ed",
    "5nn-ed",
    ["kneighborstimeseriesregressor", "1nn-dtw"],
    "5nn-dtw",
    "1nn-msm",
    "5nn-msm",
]
feature_based_regressors = [
    "summary-500",
    ["summaryregressor", "summary"],
    "catch22-500",
    ["catch22regressor", "catch22"],
    ["freshprinceregressor", "fresh-prince", "freshprince"],
    "freshprince-500",
    "tsfresh-nofs",
    ["tsfreshregressor", "tsfresh"],
    ["fpcaregressor", "fpcregressor", "fpcr"],
    ["fpcar-b-spline", "fpcr-b-spline", "fpcr-bs"],
]
hybrid_regressors = [
    ["ristregressor", "rist", "rist-extrat"],
]
interval_based_regressors = [
    ["timeseriesforestregressor", "tsf"],
    "tsf-i",
    "tsf-500",
    "rise-500",
    ["randomintervalspectralensembleregressor", "rise"],
    "cif-500",
    ["canonicalintervalforestregressor", "cif"],
    ["drcif", "drcifregressor"],
    "drcif-500",
    "summary-intervals",
    ["randomintervals-500", "catch22-intervals-500"],
    ["randomintervalregressor", "randomintervals", "catch22-intervals"],
]
other_regressors = [
    ["dummyregressor", "dummy", "dummyregressor-tsml"],
    "dummyregressor-aeon",
    ["dummyregressor-sklearn", "meanpredictorregressor", "dummymeanpred"],
    ["medianpredictorregressor", "dummymedianpred"],
]
shapelet_based_regressors = [
    ["rdstregressor", "rdst"],
]
vector_regressors = [
    ["rotationforestregressor", "rotf", "rotationforest"],
    ["linearregression", "lr"],
    ["ridgecv", "ridge"],
    ["svr", "svm", "supportvectorregressor"],
    ["grid-svr", "grid-svm", "grid-supportvectorregressor"],
    ["randomforestregressor", "rf", "randomforest"],
    ["randomforest-500", "rf-500"],
    ["xgbregressor", "xgboost"],
    ["xgb-100", "xgboost-100"],
    ["xgb-500", "xgboost-500"],
]


def get_regressor_by_name(
    regressor_name,
    random_state=None,
    n_jobs=1,
    fit_contract=0,
    checkpoint=None,
    **kwargs,
):
    """Return a regressor matching a given input name.

    Basic way of creating a regressor to build using the default or alternative
    settings. This set up is to help with batch jobs for multiple problems and to
    facilitate easy reproducibility through run_regression_experiment.

    Generally, inputting a regressor class name will return said regressor with
    default settings.

    Parameters
    ----------
    regressor_name : str
        String indicating which regressor to be returned.
    random_state : int, RandomState instance or None, default=None
        Random seed or RandomState object to be used in the regressor if available.
    n_jobs: int, default=1
        The number of jobs to run in parallel for both regressor ``fit`` and
        ``predict`` if available. `-1` means using all processors.
    fit_contract: int, default=0
        Contract time in minutes for regressor ``fit`` if available.
    checkpoint: str or None, default=None
        Path to a checkpoint file to save the regressor if available. No checkpointing
        if None.
    **kwargs
        Additional keyword arguments to be passed to the regressor.

    Return
    ------
    regressor: A BaseRegressor.
        The regressor matching the input regressor name.
    """
    r = regressor_name.lower()

    if str_in_nested_list(convolution_based_regressors, r):
        return _set_regressor_convolution_based(
            r, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(deep_learning_regressors, r):
        return _set_regressor_deep_learning(
            r, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(distance_based_regressors, r):
        return _set_regressor_distance_based(
            r, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(feature_based_regressors, r):
        return _set_regressor_feature_based(
            r, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(hybrid_regressors, r):
        return _set_regressor_hybrid(
            r, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(interval_based_regressors, r):
        return _set_regressor_interval_based(
            r, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(other_regressors, r):
        return _set_regressor_other(
            r, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(shapelet_based_regressors, r):
        return _set_regressor_shapelet_based(
            r, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    elif str_in_nested_list(vector_regressors, r):
        return _set_regressor_vector(
            r, random_state, n_jobs, fit_contract, checkpoint, kwargs
        )
    else:
        raise ValueError(f"UNKNOWN REGRESSOR: {r} in set_regressor")


def _set_regressor_convolution_based(
    r, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if r == "rocketregressor" or r == "rocket":
        from aeon.regression.convolution_based import RocketRegressor

        return RocketRegressor(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif r == "minirocket" or r == "minirocketregressor":
        from aeon.regression.convolution_based import RocketRegressor

        return RocketRegressor(
            rocket_transform="minirocket",
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "multirocket" or r == "multirocketregressor":
        from aeon.regression.convolution_based import RocketRegressor

        return RocketRegressor(
            rocket_transform="multirocket",
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "hydraregressor" or r == "hydra":
        from aeon.regression.convolution_based import HydraRegressor

        return HydraRegressor(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif (
        r == "multirockethydraregressor"
        or r == "multirockethydra"
        or r == "multirocket-hydra"
    ):
        from aeon.regression.convolution_based import MultiRocketHydraRegressor

        return MultiRocketHydraRegressor(
            random_state=random_state, n_jobs=n_jobs, **kwargs
        )


def _set_regressor_deep_learning(
    r, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if r == "cnnregressor" or r == "cnn":
        from aeon.regression.deep_learning import CNNRegressor

        return CNNRegressor(random_state=random_state, **kwargs)
    elif r == "fcnregressor" or r == "fcnn" or r == "fcn":
        from aeon.regression.deep_learning import FCNRegressor

        return FCNRegressor(random_state=random_state, **kwargs)
    elif r == "mlpregressor" or r == "mlp":
        from aeon.regression.deep_learning import MLPRegressor

        return MLPRegressor(random_state=random_state, **kwargs)
    elif r == "encoderregressor" or r == "encoder":
        from aeon.regression.deep_learning import EncoderRegressor

        return EncoderRegressor(random_state=random_state, **kwargs)
    elif r == "tapnetregressor" or r == "tapnet":
        from aeon.regression.deep_learning import TapNetRegressor

        return TapNetRegressor(random_state=random_state, **kwargs)
    elif r == "resnetregressor" or r == "resnet":
        from aeon.regression.deep_learning import ResNetRegressor

        return ResNetRegressor(random_state=random_state, **kwargs)
    elif (
        r == "individualinceptionregressor"
        or r == "singleinception"
        or r == "individualinception"
    ):
        from aeon.regression.deep_learning import IndividualInceptionRegressor

        return IndividualInceptionRegressor(random_state=random_state, **kwargs)
    elif r == "inceptiontimeregressor" or r == "inception" or r == "inceptiontime":
        from aeon.regression.deep_learning import InceptionTimeRegressor

        return InceptionTimeRegressor(random_state=random_state, **kwargs)

    elif r == "h-inceptiontimeregressor" or r == "h-inceptiontime":
        from aeon.regression.deep_learning import InceptionTimeRegressor

        return InceptionTimeRegressor(
            use_custom_filters=True, random_state=random_state, **kwargs
        )
    elif r == "litetimeregressor" or r == "litetime":
        from aeon.regression.deep_learning import LITETimeRegressor

        return LITETimeRegressor(random_state=random_state, **kwargs)


def _set_regressor_distance_based(
    r, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if r == "1nn-ed":
        from aeon.regression.distance_based import KNeighborsTimeSeriesRegressor

        return KNeighborsTimeSeriesRegressor(**kwargs)
    elif r == "5nn-ed":
        from aeon.regression.distance_based import KNeighborsTimeSeriesRegressor

        return KNeighborsTimeSeriesRegressor(n_neighbors=5, **kwargs)
    elif r == "kneighborstimeseriesregressor" or r == "1nn-dtw":
        from aeon.regression.distance_based import KNeighborsTimeSeriesRegressor

        return KNeighborsTimeSeriesRegressor(
            distance="dtw", distance_params={"window": 0.1}, **kwargs
        )
    elif r == "5nn-dtw":
        from aeon.regression.distance_based import KNeighborsTimeSeriesRegressor

        return KNeighborsTimeSeriesRegressor(
            n_neighbors=5, distance="dtw", distance_params={"window": 0.1}, **kwargs
        )
    elif r == "1nn-msm":
        from aeon.regression.distance_based import KNeighborsTimeSeriesRegressor

        return KNeighborsTimeSeriesRegressor(
            n_neighbors=1,
            distance="msm",
            distance_params={"window": None, "independent": True, "c": 1},
        )
    elif r == "5nn-msm":
        from aeon.regression.distance_based import KNeighborsTimeSeriesRegressor

        return KNeighborsTimeSeriesRegressor(
            n_neighbors=5,
            distance="msm",
            distance_params={"window": None, "independent": True, "c": 1},
        )


def _set_regressor_feature_based(
    r, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if r == "summary-500":
        from aeon.regression.feature_based import SummaryRegressor
        from sklearn.ensemble import RandomForestRegressor

        return SummaryRegressor(
            estimator=RandomForestRegressor(n_estimators=500),
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "summaryregressor" or r == "summary":
        from aeon.regression.feature_based import SummaryRegressor

        return SummaryRegressor(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif r == "catch22-500":
        from aeon.regression.feature_based import Catch22Regressor
        from sklearn.ensemble import RandomForestRegressor

        return Catch22Regressor(
            estimator=RandomForestRegressor(n_estimators=500),
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "catch22regressor" or r == "catch22":
        from aeon.regression.feature_based import Catch22Regressor

        return Catch22Regressor(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif r == "freshprinceregressor" or r == "fresh-prince" or r == "freshprince":
        from aeon.regression.feature_based import FreshPRINCERegressor

        return FreshPRINCERegressor(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "freshprince-500":
        from aeon.regression.feature_based import FreshPRINCERegressor

        return FreshPRINCERegressor(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "tsfresh-nofs":
        from aeon.regression.feature_based import TSFreshRegressor

        return TSFreshRegressor(
            relevant_feature_extractor=False,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "tsfreshregressor" or r == "tsfresh":
        from aeon.regression.feature_based import TSFreshRegressor

        return TSFreshRegressor(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif r == "fpcaregressor" or r == "fpcregressor" or r == "fpcr":
        from tsml.feature_based import FPCARegressor

        return FPCARegressor(n_jobs=n_jobs, **kwargs)
    elif r == "fpcar-b-spline" or r == "fpcr-b-spline" or r == "fpcr-bs":
        from tsml.feature_based import FPCARegressor

        return FPCARegressor(n_jobs=n_jobs, bspline=True, order=4, n_basis=10, **kwargs)


def _set_regressor_hybrid(r, random_state, n_jobs, fit_contract, checkpoint, kwargs):
    if r == "ristregressor" or r == "rist" or r == "rist-extrat":
        from aeon.regression.hybrid import RISTRegressor
        from sklearn.ensemble import ExtraTreesRegressor

        return RISTRegressor(
            random_state=random_state,
            n_jobs=n_jobs,
            estimator=ExtraTreesRegressor(n_estimators=500),
            **kwargs,
        )


def _set_regressor_interval_based(
    r, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if r == "timeseriesforestregressor" or r == "tsf":
        from aeon.regression.interval_based import TimeSeriesForestRegressor

        return TimeSeriesForestRegressor(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "tsf-i":
        from aeon.regression.interval_based import TimeSeriesForestRegressor
        from tsml.compose import ChannelEnsembleRegressor

        estimators = [
            (
                "tsf",
                TimeSeriesForestRegressor(
                    random_state=random_state,
                    n_jobs=n_jobs,
                ),
                "all-split",
            )
        ]

        return ChannelEnsembleRegressor(estimators, **kwargs)
    elif r == "tsf-500":
        from aeon.regression.interval_based import TimeSeriesForestRegressor

        return TimeSeriesForestRegressor(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "rise-500":
        from aeon.regression.interval_based import (
            RandomIntervalSpectralEnsembleRegressor,
        )

        return RandomIntervalSpectralEnsembleRegressor(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "randomintervalspectralensembleregressor" or r == "rise":
        from aeon.regression.interval_based import (
            RandomIntervalSpectralEnsembleRegressor,
        )

        return RandomIntervalSpectralEnsembleRegressor(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "cif-500":
        from aeon.regression.interval_based import CanonicalIntervalForestRegressor

        return CanonicalIntervalForestRegressor(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "canonicalintervalforestregressor" or r == "cif":
        from aeon.regression.interval_based import CanonicalIntervalForestRegressor

        return CanonicalIntervalForestRegressor(
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "drcif" or r == "drcifregressor":
        from aeon.regression.interval_based import DrCIFRegressor

        return DrCIFRegressor(
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif r == "drcif-500":
        from aeon.regression.interval_based import DrCIFRegressor

        return DrCIFRegressor(
            n_estimators=500,
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif r == "summary-intervals":
        from aeon.regression.interval_based import RandomIntervalRegressor
        from aeon.transformations.collection.feature_based import (
            SevenNumberSummaryTransformer,
        )
        from sklearn.ensemble import RandomForestRegressor

        return RandomIntervalRegressor(
            features=SevenNumberSummaryTransformer(),
            estimator=RandomForestRegressor(n_estimators=500),
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "randomintervals-500" or r == "catch22-intervals-500":
        from aeon.regression.interval_based import RandomIntervalRegressor
        from sklearn.ensemble import RandomForestRegressor

        return RandomIntervalRegressor(
            estimator=RandomForestRegressor(n_estimators=500),
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif (
        r == "randomintervalregressor"
        or r == "randomintervals"
        or r == "catch22-intervals"
    ):
        from aeon.regression.interval_based import RandomIntervalRegressor

        return RandomIntervalRegressor(
            random_state=random_state, n_jobs=n_jobs, **kwargs
        )


def _set_regressor_other(r, random_state, n_jobs, fit_contract, checkpoint, kwargs):
    if r == "dummy" or r == "dummyregressor" or r == "dummyregressor-tsml":
        from tsml.dummy import DummyRegressor

        return DummyRegressor(**kwargs)
    elif r == "dummyregressor-aeon":
        from aeon.regression import DummyRegressor

        return DummyRegressor(**kwargs)
    elif (
        r == "dummyregressor-sklearn"
        or r == "meanpredictorregressor"
        or r == "dummymeanpred"
    ):
        from sklearn.dummy import DummyRegressor

        return DummyRegressor(**kwargs)
    elif r == "medianpredictorregressor" or r == "dummymedianpred":
        from sklearn.dummy import DummyRegressor

        return DummyRegressor(strategy="median", **kwargs)


def _set_regressor_shapelet_based(
    r, random_state, n_jobs, fit_contract, checkpoint, kwargs
):
    if r == "rdstregressor" or r == "rdst":
        from aeon.regression.shapelet_based import RDSTRegressor

        return RDSTRegressor(random_state=random_state, n_jobs=n_jobs, **kwargs)


def _set_regressor_vector(r, random_state, n_jobs, fit_contract, checkpoint, kwargs):
    if r == "rotationforestregressor" or r == "rotf" or r == "rotationforest":
        from aeon.regression.sklearn import RotationForestRegressor

        return RotationForestRegressor(
            random_state=random_state,
            n_jobs=n_jobs,
            time_limit_in_minutes=fit_contract,
            **kwargs,
        )
    elif r == "linearregression" or r == "lr":
        from sklearn.linear_model import LinearRegression

        return LinearRegression(fit_intercept=True, n_jobs=n_jobs, **kwargs)
    elif r == "ridgecv" or r == "ridge":
        from sklearn.linear_model import RidgeCV

        return RidgeCV(fit_intercept=True, alphas=np.logspace(-3, 3, 10), **kwargs)
    elif r == "svr" or r == "svm" or r == "supportvectorregressor":
        from sklearn.svm import SVR

        return SVR(kernel="rbf", C=1, **kwargs)
    elif r == "grid-svr" or r == "grid-svm" or r == "grid-supportvectorregressor":
        from sklearn.model_selection import GridSearchCV
        from sklearn.svm import SVR

        param_grid = [
            {
                "kernel": ["rbf", "sigmoid"],
                "C": [0.1, 1, 10, 100],
                "gamma": [0.001, 0.01, 0.1, 1],
            }
        ]

        return GridSearchCV(
            estimator=SVR(),
            param_grid=param_grid,
            scoring="neg_mean_squared_error",
            n_jobs=n_jobs,
            cv=3,
            **kwargs,
        )
    elif r == "randomforestregressor" or r == "rf" or r == "randomforest":
        from sklearn.ensemble import RandomForestRegressor

        return RandomForestRegressor(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif r == "rf-500" or r == "randomforest-500":
        from sklearn.ensemble import RandomForestRegressor

        return RandomForestRegressor(
            n_estimators=500, random_state=random_state, n_jobs=n_jobs, **kwargs
        )
    elif r == "xgbregressor" or r == "xgboost":
        from xgboost import XGBRegressor

        return XGBRegressor(random_state=random_state, n_jobs=n_jobs, **kwargs)
    elif r == "xgb-100" or r == "xgboost-100":
        from xgboost import XGBRegressor

        return XGBRegressor(
            n_estimators=100,
            learning_rate=0.1,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )
    elif r == "xgb-500" or r == "xgboost-500":
        from xgboost import XGBRegressor

        return XGBRegressor(
            n_estimators=500,
            learning_rate=0.1,
            random_state=random_state,
            n_jobs=n_jobs,
            **kwargs,
        )



import numpy as np
from numba import njit, int64, prange
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils.validation import check_random_state


def ps(x, sign=1, n=None, pad="mean"):
    """Power spectrum transformer.

    Power spectrum transform, currently calculated using np function.
    It would be worth looking at ff implementation, see difference in speed
    to java.

    Parameters
    ----------
    x : array-like shape = [interval_width]
    sign : {-1, 1}, default = 1
    n : int, default=None
    pad : str or function, default='mean'
        controls the mode of the pad function
        see numpy.pad for more details
        https://numpy.org/doc/stable/reference/generated/numpy.pad.html

    Returns
    -------
    y : array-like shape = [len(x)/2]
    """
    x_len = x.shape[-1]
    x_is_1d = x.ndim == 1
    # pad or slice series if length is not of power of 2 or n is specified
    if x_len & (x_len - 1) != 0 or n:
        # round n (or the length of x) to next power of 2
        # when n is not specified
        if not n:
            n = _round_to_nearest_power_of_two(x_len)
        # pad series up to n when n is larger otherwise slice series up to n
        if n > x_len:
            pad_length = (0, n - x_len) if x_is_1d else ((0, 0), (0, n - x_len))
            x_in_power_2 = np.pad(x, pad_length, mode=pad)
        else:
            x_in_power_2 = x[:n] if x_is_1d else x[:, :n]
    else:
        x_in_power_2 = x
    # use sign to determine inverse or normal fft
    # using the norm in numpy fft function
    # backward = normal fft, forward = inverse fft (divide by n after fft)
    # note: use the following code when upgrade numpy to 1.20
    # norm = "backward" if sign > 0 else "forward"
    # fft = np.fft.rfft(x_in_power_2, norm=norm)
    if sign < 0:
        x_in_power_2 /= n
    fft = np.fft.rfft(x_in_power_2)
    fft = fft[:-1] if x_is_1d else fft[:, :-1]
    return np.abs(fft)


@njit("int64(int64)", cache=True)
def _round_to_nearest_power_of_two(n):
    return int64(1 << round(np.log2(n)))


@njit(parallel=True, cache=True)
def acf(x, max_lag):
    """Autocorrelation function transform.

    currently calculated using standard stats method. We could use inverse of power
    spectrum, especially given we already have found it, worth testing for speed and
    correctness. HOWEVER, for long series, it may not give much benefit, as we do not
    use that many ACF terms.

    Parameters
    ----------
    x : array-like shape = [interval_width]
    max_lag: int
        The number of ACF terms to find.

    Returns
    -------
    y : array-like shape = [max_lag]
    """
    y = np.empty(max_lag)
    length = len(x)
    for lag in prange(1, max_lag + 1):
        # Do it ourselves to avoid zero variance warnings
        lag_length = length - lag
        x1, x2 = x[:-lag], x[lag:]
        s1 = np.sum(x1)
        s2 = np.sum(x2)
        m1 = s1 / lag_length
        m2 = s2 / lag_length
        ss1 = np.sum(x1 * x1)
        ss2 = np.sum(x2 * x2)
        v1 = ss1 - s1 * m1
        v2 = ss2 - s2 * m2
        v1_is_zero, v2_is_zero = v1 <= 1e-9, v2 <= 1e-9
        if v1_is_zero and v2_is_zero:  # Both zero variance,
            # so must be 100% correlated
            y[lag - 1] = 1
        elif v1_is_zero or v2_is_zero:  # One zero variance
            # the other not
            y[lag - 1] = 0
        else:
            y[lag - 1] = np.sum((x1 - m1) * (x2 - m2)) / np.sqrt(v1 * v2)
        # _x = np.vstack((x[:-lag], x[lag:]))
        # s = np.sum(_x, axis=1)
        # ss = np.sum(_x * _x, axis=1)
        # v = ss - s * s / l
        # zero_variances = v <= 1e-9
        # i = lag - 1
        # if np.all(zero_variances):  # Both zero variance,
        #     # so must be 100% correlated
        #     y[i] = 1
        # elif np.any(zero_variances):  # One zero variance
        #     # the other not
        #     y[i] = 0
        # else:
        #     m = _x - s.reshape(2, 1) / l
        #     y[i] = (m[0] @ m[1]) / np.sqrt(np.prod(v))

    return y

def _acf(X, istart, iend, lag):
    n_instances, _ = X.shape
    acf_x = np.empty(shape=(n_instances, lag))
    for j in range(n_instances):
        interval_x = X[j, istart:iend]
        acf_x[j] = acf(interval_x, lag)

    return acf_x


def _ps(X, istart, iend, lag):
    n_instances, _ = X.shape
    ps_len = _round_to_nearest_power_of_two(istart - iend)
    ps_x = np.empty(shape=(n_instances, ps_len))
    for j in range(n_instances):
        interval_x = X[j, istart:iend]
        ps_x[j] = ps(interval_x, n=ps_len * 2)

    return ps_x


FEATURE_CANDIDATES = [_acf, _ps]


class IntervalSplitter:
    """RISE-based splitter for TS-CHIEF implementation."""

    @staticmethod
    def generate(X, y, random_state=None):
        """Generate a randomized interval splitter candidate."""
        samples, dims, length = X.shape
        splitter = IntervalSplitter()
        splitter.rng = check_random_state(random_state)
        splitter.dim = splitter.rng.randint(dims)

        min_interval = min(16, length)
        acf_min_values = 4

        splitter.istart = splitter.rng.randint(0, length - min_interval)
        splitter.iend = splitter.rng.randint(splitter.istart + min_interval, length)
        acf_lag = 100
        if acf_lag > splitter.iend - splitter.istart - acf_min_values:
            acf_lag = splitter.iend - splitter.istart - acf_min_values
        if acf_lag < 0:
            acf_lag = 1

        splitter.acf_lag = acf_lag

        splitter.transform = splitter.rng.choice(FEATURE_CANDIDATES)
        X_transformed = splitter.transform(
            X[:, splitter.dim, :], splitter.istart, splitter.iend, splitter.acf_lag
        )

        splitter.tree = DecisionTreeClassifier(
            criterion="gini", max_depth=1, max_features=None, random_state=splitter.rng
        ).fit(X_transformed, y)

        return splitter

    def split(self, X):
        """Split incoming data."""
        X = X[:, self.dim, :]
        X_transformed = self.transform(X, self.istart, self.iend, self.acf_lag)

        return self.tree.apply(X_transformed) - 1



__all__ = [
    "DictionarySplitter",
    "generate_boss_transforms",
    "IntervalSplitter",
    "DistanceSplitter",
]

from tsml_eval._wip.tschief._splitters.dictionary_splitter import (
    DictionarySplitter,
    generate_boss_transforms,
)
from tsml_eval._wip.tschief._splitters.distance_splitter import DistanceSplitter
from tsml_eval._wip.tschief._splitters.interval_splitter import IntervalSplitter



from itertools import product

import numpy as np
from sklearn.utils.validation import check_random_state
from aeon.classification.dictionary_based._boss import boss_distance
from aeon.transformations.collection.dictionary_based import SFAFast


class DictionarySplitter:
    """BOSS-based splitter for TS-CHIEF implementation."""

    @staticmethod
    def generate(X_boss, y, sfas, random_state=None):
        """Generate a randomized dictionary splitter candidate."""
        dims, num_transforms = X_boss.shape
        rng = check_random_state(random_state)

        splitter = DictionarySplitter()
        splitter.dim = rng.randint(dims)
        splitter.transform_idx = rng.randint(num_transforms)
        splitter.sfa = sfas[splitter.dim, splitter.transform_idx]

        splitter.exemplars = []
        classes = np.unique(y)
        for c in classes:
            group = np.argwhere(y == c).ravel()
            exemplar_idx = rng.choice(group)
            splitter.exemplars.append(
                X_boss[splitter.dim, splitter.transform_idx][exemplar_idx, :]
            )

        return splitter

    def split_train(self, X_boss):
        """Split the training data without needlessly SFAing again."""
        X_boss = X_boss[self.dim, self.transform_idx]
        samples, _ = X_boss.shape

        split_idx = np.empty(samples, dtype=int)
        for i in range(samples):
            distances = [
                boss_distance(X_boss[i, :], exemplar, 0) for exemplar in self.exemplars
            ]
            split_idx[i] = np.argmin(distances)

        return split_idx

    def split(self, X):
        """Split incoming data."""
        X = X[:, [self.dim], :]
        X_boss = self.sfa.transform(X)
        samples, _ = X_boss.shape

        split_idx = np.empty(samples, dtype=int)
        for i in range(samples):
            distances = [
                boss_distance(X_boss[i, :], exemplar, 0) for exemplar in self.exemplars
            ]
            split_idx[i] = np.argmin(distances)

        return split_idx


def generate_boss_transforms(X, num_transforms_per_dim=1000, random_state=None):
    """Generate random SFA transformations for splitters."""
    rng = check_random_state(random_state)
    _, dims, length = X.shape

    window_sizes = np.arange(10, length + 1)
    word_lengths = np.array([6, 8, 10, 12, 14, 16])
    normalizations = np.array([True, False])
    alphabet_sizes = np.array([4])

    params = list(product(window_sizes, word_lengths, normalizations, alphabet_sizes))
    params = np.array(params, dtype=tuple)
    num_transforms_per_dim = min(num_transforms_per_dim, len(params))

    sfas = []
    bags = []

    for dim in range(dims):
        sfas_dim = []
        bags_dim = []

        used_params_idx = rng.choice(
            len(params), size=num_transforms_per_dim, replace=False
        )
        used_params = params[used_params_idx]
        for window_size, word_length, norm, alphabet_size in used_params:
            sfa = SFAFast(
                word_length=word_length,
                alphabet_size=alphabet_size,
                window_size=window_size,
                norm=norm,
                binning_method="equi-depth",
                anova=False,
                variance=False,
                bigrams=False,
                skip_grams=False,
                remove_repeat_words=True,
                feature_selection="none",
                random_state=rng,
                n_jobs=1,
            )
            bag = sfa.fit_transform(X[:, [dim], :])

            sfas_dim.append(sfa)
            bags_dim.append(bag)

        sfas.append(sfas_dim)
        bags.append(bags_dim)

    return np.array(sfas), np.array(bags)



"""Tests for publication experiments estimator selection."""

import pytest

from tsml_eval.publications.y2023.distance_based_clustering import (
    _set_distance_clusterer,
    distance_based_clusterers,
)
from tsml_eval.testing.testing_utils import _check_set_method, _check_set_method_results


def test_set_distance_clusterer():
    """Test set_distance_clusterer method."""
    clusterer_dict = {}
    all_clusterer_names = []

    _check_set_method(
        _set_distance_clusterer,
        distance_based_clusterers,
        clusterer_dict,
        all_clusterer_names,
    )

    _check_set_method_results(
        clusterer_dict,
        estimator_name="Clusterers",
        method_name="_set_distance_clusterer",
    )


def test_set_distance_clusterer_invalid():
    """Test set_distance_clusterer method with invalid estimator."""
    with pytest.raises(ValueError, match="UNKNOWN CLUSTERER"):
        _set_distance_clusterer("invalid")



"""Storage for estimator results and result i/o."""

__all__ = [
    "ClassifierResults",
    "ClustererResults",
    "ForecasterResults",
    "RegressorResults",
    "load_classifier_results",
    "load_clusterer_results",
    "load_forecaster_results",
    "load_regressor_results",
]

from tsml_eval.evaluation.storage.classifier_results import (
    ClassifierResults,
    load_classifier_results,
)
from tsml_eval.evaluation.storage.clusterer_results import (
    ClustererResults,
    load_clusterer_results,
)
from tsml_eval.evaluation.storage.forecaster_results import (
    ForecasterResults,
    load_forecaster_results,
)
from tsml_eval.evaluation.storage.regressor_results import (
    RegressorResults,
    load_regressor_results,
)
